from __future__ import absolute_import
import numpy


def princomp(x):
    """Determine the principal components of a vector of measurements
    
    Determine the principal components of a vector of measurements
    x should be a M x N numpy array composed of M observations of n variables
    The output is:
    coeffs - the NxN correlation matrix that can be used to transform x into its components
    
    The code for this function is based on "A Tutorial on Principal Component
    Analysis", Shlens, 2005 http://www.snl.salk.edu/~shlens/pub/notes/pca.pdf
    (unpublished)
    """

    (M, N) = x.shape
    Mean = x.mean(0)
    y = x - Mean
    cov = numpy.dot(y.transpose(), y) / (M - 1)
    (V, PC) = numpy.linalg.eig(cov)
    order = (-V).argsort()
    coeff = PC[:, order]
    return coeff



from __future__ import absolute_import
import skimage.segmentation


def watershed(image, markers, connectivity=None, offset=None, mask=None):
    return skimage.segmentation.watershed(image, markers, connectivity, offset, mask)



from __future__ import absolute_import
from __future__ import print_function
import warnings

import numpy as np
from scipy import linspace
from scipy.ndimage import affine_transform
from six.moves import range

"""Automatically determine whether background is darker than foreground"""
MODE_AUTO = "auto"

"""Background is darker than foreground"""
MODE_DARK = "dark"

"""Background is brighter than foreground"""
MODE_BRIGHT = "bright"

"""Some foreground is darker, some is lighter"""
MODE_GRAY = "gray"


def prcntiles(x, percents):
    """Equivalent to matlab prctile(x,p), uses linear interpolation."""
    x = np.array(x).flatten()
    listx = np.sort(x)
    xpcts = []
    lenlistx = len(listx)
    refs = []
    for i in range(0, lenlistx):
        r = 100 * (
            (0.5 + i) / lenlistx
        )  # refs[i] is percentile of listx[i] in matrix x
        refs.append(r)

    rpcts = []
    for p in percents:
        if p < refs[0]:
            rpcts.append(listx[0])
        elif p > refs[-1]:
            rpcts.append(listx[-1])
        else:
            for j in range(0, lenlistx):  # lenlistx=len(refs)
                if refs[j] <= p and refs[j + 1] >= p:
                    my = listx[j + 1] - listx[j]
                    mx = refs[j + 1] - refs[j]
                    m = my / mx  # slope of line between points
                    rpcts.append((m * (p - refs[j])) + listx[j])
                    break
    xpcts.append(rpcts)
    return np.array(xpcts).transpose()


def automode(data):
    """Tries to guess if the image contains dark objects on a bright background (1)
or if the image contains bright objects on a dark background (-1),
or if it contains both dark and bright objects on a gray background (0)."""

    pct = prcntiles(np.array(data), [1, 20, 80, 99])

    upper = pct[3] - pct[2]
    mid = pct[2] - pct[1]
    lower = pct[1] - pct[0]

    ##print 'upper = '+str(upper)
    ##print 'mid = '+str(mid)
    ##print 'lower = '+str(lower)

    # upper objects
    if upper > mid:
        uo = 1
    else:
        uo = 0
    ##print 'uo = '+str(uo)

    # lower objects
    if lower > mid:
        lo = 1
    else:
        lo = 0
    ##print 'lo = '+str(lo)

    if uo == 1:
        if lo == 1:
            mode = 0
            # both upper and lower objects
        else:
            mode = -1
            # only upper objects
    else:
        if lo == 1:
            mode = 1
            # only lower objects
        else:
            mode = 0
            # no objects at all

    return mode


def spline_factors(u):
    """u is np.array"""

    X = np.array(
        [
            (1.0 - u) ** 3,
            4 - (6.0 * (u ** 2)) + (3.0 * (u ** 3)),
            1.0 + (3.0 * u) + (3.0 * (u ** 2)) - (3.0 * (u ** 3)),
            u ** 3,
        ]
    ) * (1.0 / 6)

    return X


def pick(picklist, val):
    """Index to first value in picklist that is larger than val.
If none is larger, index=len(picklist)."""

    assert np.all(np.sort(picklist) == picklist), "pick list is not ordered correctly"
    val = np.array(val)
    i_pick, i_val = np.mgrid[0 : len(picklist), 0 : len(val)]
    #
    # Mark a picklist entry as 1 if the value is before or at,
    # mark it as zero if it is afterward
    #
    is_not_larger = picklist[i_pick] <= val[i_val]
    #
    # The index is the number of entries that are 1
    #
    p = np.sum(is_not_larger, 0)

    return p


def confine(x, low, high):
    """Confine x to [low,high]. Values outside are set to low/high.
See also restrict."""

    y = x.copy()
    y[y < low] = low
    y[y > high] = high
    return y


def gauss(x, m_y, sigma):
    """returns the gaussian with mean m_y and std. dev. sigma,
calculated at the points of x."""

    e_y = [
        np.exp((1.0 / (2 * float(sigma) ** 2) * -((n - m_y) ** 2))) for n in np.array(x)
    ]
    y = [1.0 / (float(sigma) * np.sqrt(2 * np.pi)) * e for e in e_y]

    return np.array(y)


def d2gauss(x, m_y, sigma):
    """returns the second derivative of the gaussian with mean m_y,
and standard deviation sigma, calculated at the points of x."""

    return gauss(x, m_y, sigma) * [
        -1 / sigma ** 2 + (n - m_y) ** 2 / sigma ** 4 for n in x
    ]


def spline_matrix(x, px):
    n = len(px)
    lx = len(x)

    # Assign each x to an interval.  Subtract 1 to get the beginning of the interval.
    px = np.array(px)
    x = np.array(x)
    j = np.array(pick(px, x)) - 1
    #
    # We need at least one entry before and two after for the four factors
    # of the cubic spline
    #
    j = confine(j, 1, n - 3)

    u = (x - px[j]) / (
        px[j + 1] - px[j]
    )  # how far are we on the line segment px[j]->px[j+1], 0<=u<1

    spf = spline_factors(u)
    #
    # Set up to broadcast spf to the correct spline factors
    # The cubic has four factors that broadcast starting at j-1 to j+2
    #
    ii, jj = np.mgrid[0 : spf.shape[0], 0:lx]
    V = np.zeros((n, lx))
    V[j[jj] - 1 + ii, jj] = spf[ii, jj]
    return V


def spline_matrix2d(x, y, px, py, mask=None):
    """For boundary constraints, the first two and last two spline pieces are constrained
to be part of the same cubic curve."""
    V = np.kron(spline_matrix(x, px), spline_matrix(y, py))

    lenV = len(V)

    if mask is not None:
        indices = np.nonzero(mask.T.flatten())
        if len(indices) > 1:
            indices = np.nonzero(mask.T.flatten())[1][0]
        newV = V.T[indices]
        V = newV.T

        V = V.reshape((V.shape[0], V.shape[1]))

    return V


def splinefit2d(x, y, z, px, py, mask=None):
    """Make a least squares fit of the spline (px,py,pz) to the surface (x,y,z).
If mask is given, only masked points are used for the regression."""

    if mask is None:
        V = np.array(spline_matrix2d(x, y, px, py))
        a = np.array(z.T.flatten())
        pz = np.linalg.lstsq(V.T, a.T)[0].T
    else:
        V = np.array(spline_matrix2d(x, y, px, py, mask))
        indices = np.nonzero(np.array(mask).T.flatten())
        if len(indices[0]) == 0:
            pz = np.zeros((len(py), len(px)))
        # indices is empty when mask changes to all zeros
        else:
            a = np.array((z.T.flatten()[indices[0]]))
            pz = np.linalg.lstsq(V.T, a.T)[0].T

    pz = pz.reshape((len(py), len(px)))
    return pz.transpose()


def splineimage(Z, points, mask=None, x=None, y=None):

    [r, c] = Z.shape
    if x is None:
        x = np.arange(c)
    if y is None:
        y = np.arange(r)

    # spline control points, let the outer ones be outside the image
    cstep = (x[-1] - x[0]) / (points - 3.0)
    px = linspace(x[0] - cstep, x[-1] + cstep, points)
    rstep = (y[-1] - y[0]) / (points - 3.0)
    py = linspace(y[0] - rstep, y[-1] + rstep, points)

    if mask is None:
        pz = splinefit2d(x, y, Z, px, py)
    else:
        pz = splinefit2d(x, y, Z, px, py, mask)
    return px, py, pz


def evalspline2d(x, y, px, py, pz):

    V = spline_matrix2d(x, y, px, py)
    a = np.dot(pz.T.reshape((1, np.prod(pz.shape))), V)
    z = a.reshape((len(x), len(y))).T

    return z


def unbiased_std(a):
    return np.sqrt(((a - np.mean(a)) ** 2).sum() / (len(a) - 1))


def backgr(
    img,
    mask=None,
    mode=MODE_AUTO,
    thresh=2,
    splinepoints=5,
    scale=1,
    maxiter=40,
    convergence=0.001,
):
    """Iterative spline-based background correction.
    
    mode - one of MODE_AUTO, MODE_DARK, MODE_BRIGHT or MODE_GRAY
    thresh - thresh is threshold to cut at, in units of sigma. 
    splinepoints - # of points in spline in each direction
    scale - scale the image by this factor (e.g. 2 = operate on 1/2 of the points)
    maxiter - maximum # of iterations
    convergence - result has converged when the standard deviation of the
                  difference between iterations is this fraction of the
                  maximum image intensity.
Spline mesh is splinepoints x splinepoints. Modes are
defined by the background intensity, Larger thresh->slower but
more stable convergence. Returns background matrix."""

    assert img.ndim == 2, "Image must be 2-d"
    assert splinepoints >= 3, "The minimum grid size is 3x3"
    assert maxiter >= 1
    assert mode in [MODE_AUTO, MODE_BRIGHT, MODE_DARK, MODE_GRAY], (
        mode + " is not a valid background mode"
    )
    orig_shape = np.array(img.shape).copy()
    input_mask = mask
    if mask is None:
        mask = np.ones(orig_shape, dtype=bool)  # start with mask = whole image
        clip_imin = clip_jmin = 0
        clip_imax = img.shape[0]
        clip_jmax = img.shape[1]
        clip_shape = orig_shape
    else:
        isum = np.sum(mask, 1)
        jsum = np.sum(mask, 0)
        clip_imin = np.min(np.argwhere(isum != 0))
        clip_imax = np.max(np.argwhere(isum != 0)) + 1
        clip_jmin = np.min(np.argwhere(jsum != 0))
        clip_jmax = np.max(np.argwhere(jsum != 0)) + 1
        clip_shape = np.array([clip_imax - clip_imin, clip_jmax - clip_jmin])

    subsample_shape = (clip_shape / scale).astype(int)
    ratio = (clip_shape.astype(float) - 1) / (subsample_shape.astype(float) - 1)
    transform = np.array([[ratio[0], 0], [0, ratio[1]]])
    inverse_transform = np.array([[1.0 / ratio[0], 0], [0, 1.0 / ratio[1]]])

    img = affine_transform(
        img[clip_imin:clip_imax, clip_jmin:clip_jmax],
        transform,
        output_shape=tuple(subsample_shape),
        order=2,
    )
    mask = (
        affine_transform(
            mask[clip_imin:clip_imax, clip_jmin:clip_jmax].astype(float),
            transform,
            output_shape=tuple(subsample_shape),
            order=2,
        )
        > 0.5
    )
    orig_mask = mask

    if mode == "auto":
        mode = automode(img[orig_mask])
    elif mode == "dark" or mode == "low":
        mode = -1
    elif mode == "bright" or mode == "high":
        mode = 1
    elif mode == "gray" or mode == "grey" or mode == "mid":
        mode = 0

    # Base the stop criterion on a fraction of the image dynamic range

    stop_criterion = max(
        (np.max(img) - np.min(img)) * convergence, np.finfo(img.dtype).eps
    )
    [r, c] = img.shape

    oldres = np.zeros((r, c))  # old background

    for i in range(maxiter):
        px, py, pz = splineimage(img, splinepoints, np.array(mask))  # now with mask
        res = evalspline2d(np.arange(c), np.arange(r), px, py, pz)
        comp = img - res

        diff = res[orig_mask] - oldres[orig_mask]
        ###Compute std. deviation in same way as matlab std(), (not numpy.std())
        stddiff = unbiased_std(diff)

        if stddiff < stop_criterion:  # stop_criterion instead of .004
            break
        elif i == maxiter:
            warnings.warn(
                "Background did not converge after %d iterations.\nMake sure that the foreground/background mode is correct."
                % (i)
            )

        oldres = res

        # calculate new mask
        backgr = comp[mask]
        sigma = unbiased_std(backgr)
        cut = sigma * thresh

        if mode < 0:
            mask = comp < cut
        elif mode > 0:
            mask = comp > -cut
        else:
            mask = abs(comp) < cut
        mask &= orig_mask
        nnz = np.sum(mask)
        if nnz < 0.01 * np.sum(orig_mask):
            warnings.warn(
                "Less than 1%% of the pixels used for fitting,\ntry starting again with a larger threshold value"
            )
            break

    output = np.zeros(orig_shape, img.dtype)
    output[clip_imin:clip_imax, clip_jmin:clip_jmax] = affine_transform(
        res, inverse_transform, output_shape=tuple(clip_shape), order=3
    )
    if input_mask is not None:
        output[~input_mask] = 0
    return output


def bg_compensate(img, sigma, splinepoints, scale):
    """Reads file, subtracts background. Returns [compensated image, background]."""

    from PIL import Image
    import pylab
    from matplotlib.image import pil_to_array
    from centrosome.filter import canny
    import matplotlib

    img = Image.open(img)
    if img.mode == "I;16":
        # 16-bit image
        # deal with the endianness explicitly... I'm not sure
        # why PIL doesn't get this right.
        imgdata = np.fromstring(img.tostring(), np.uint8)
        imgdata.shape = (int(imgdata.shape[0] / 2), 2)
        imgdata = imgdata.astype(np.uint16)
        hi, lo = (0, 1) if img.tag.prefix == "MM" else (1, 0)
        imgdata = imgdata[:, hi] * 256 + imgdata[:, lo]
        img_size = list(img.size)
        img_size.reverse()
        new_img = imgdata.reshape(img_size)
        # The magic # for maximum sample value is 281
        if 281 in img.tag:
            img = new_img.astype(np.float32) / img.tag[281][0]
        elif np.max(new_img) < 4096:
            img = new_img.astype(np.float32) / 4095.0
        else:
            img = new_img.astype(np.float32) / 65535.0
    else:
        img = pil_to_array(img)

    pylab.subplot(1, 3, 1).imshow(img, cmap=matplotlib.cm.Greys_r)
    pylab.show()

    if len(img.shape) > 2:
        raise ValueError("Image must be grayscale")

    ## Create mask that will fix problem when image has black areas outside of well
    edges = canny(img, np.ones(img.shape, bool), 2, 0.1, 0.3)
    ci = np.cumsum(edges, 0)
    cj = np.cumsum(edges, 1)
    i, j = np.mgrid[0 : img.shape[0], 0 : img.shape[1]]
    mask = ci > 0
    mask = mask & (cj > 0)
    mask[1:, :] &= ci[0:-1, :] < ci[-1, j[0:-1, :]]
    mask[:, 1:] &= cj[:, 0:-1] < cj[i[:, 0:-1], -1]

    import time

    t0 = time.clock()
    bg = backgr(img, mask, MODE_AUTO, sigma, splinepoints=splinepoints, scale=scale)
    print("Executed in %f sec" % (time.clock() - t0))
    bg[~mask] = img[~mask]

    pylab.subplot(1, 3, 2).imshow(img - bg, cmap=matplotlib.cm.Greys_r)
    pylab.subplot(1, 3, 3).imshow(bg, cmap=matplotlib.cm.Greys_r)
    pylab.show()



from __future__ import absolute_import
import numpy as np

from . import _propagate


def propagate(image, labels, mask, weight):
    """Propagate the labels to the nearest pixels
    
    image - gives the Z height when computing distance
    labels - the labeled image pixels
    mask   - only label pixels within the mask
    weight - the weighting of x/y distance vs z distance
             high numbers favor x/y, low favor z
    
    returns a label matrix and the computed distances
    """
    if image.shape != labels.shape:
        raise ValueError(
            "Image shape %s != label shape %s" % (repr(image.shape), repr(labels.shape))
        )
    if image.shape != mask.shape:
        raise ValueError(
            "Image shape %s != mask shape %s" % (repr(image.shape), repr(mask.shape))
        )
    labels_out = np.zeros(labels.shape, np.int32)
    distances = -np.ones(labels.shape, np.float64)
    distances[labels > 0] = 0
    labels_and_mask = np.logical_and(labels != 0, mask)
    coords = np.argwhere(labels_and_mask)
    i1, i2 = _propagate.convert_to_ints(0.0)
    ncoords = coords.shape[0]
    pq = np.column_stack(
        (
            np.ones((ncoords,), int) * i1,
            np.ones((ncoords,), int) * i2,
            labels[labels_and_mask],
            coords,
        )
    )
    _propagate.propagate(
        np.ascontiguousarray(image, np.float64),
        np.ascontiguousarray(pq, np.int32),
        np.ascontiguousarray(mask, np.int8),
        labels_out,
        distances,
        float(weight),
    )
    labels_out[labels > 0] = labels[labels > 0]
    return labels_out, distances



from __future__ import absolute_import
from __future__ import division
import numpy as np


def otsu(data, min_threshold=None, max_threshold=None, bins=256):
    """Compute a threshold using Otsu's method
    
    data           - an array of intensity values between zero and one
    min_threshold  - only consider thresholds above this minimum value
    max_threshold  - only consider thresholds below this maximum value
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    """
    assert (
        min_threshold is None or max_threshold is None or min_threshold < max_threshold
    )

    def constrain(threshold):
        if not min_threshold is None and threshold < min_threshold:
            threshold = min_threshold
        if not max_threshold is None and threshold > max_threshold:
            threshold = max_threshold
        return threshold

    data = np.atleast_1d(data)
    data = data[~np.isnan(data)]
    if len(data) == 0:
        return (
            min_threshold
            if not min_threshold is None
            else max_threshold
            if not max_threshold is None
            else 0
        )
    elif len(data) == 1:
        return constrain(data[0])
    if bins > len(data):
        bins = len(data)
    data.sort()
    var = running_variance(data)
    rvar = np.flipud(running_variance(np.flipud(data)))
    thresholds = data[1 : len(data) : len(data) // bins]
    score_low = var[0 : len(data) - 1 : len(data) // bins] * np.arange(
        0, len(data) - 1, len(data) // bins
    )
    score_high = rvar[1 : len(data) : len(data) // bins] * (
        len(data) - np.arange(1, len(data), len(data) // bins)
    )
    scores = score_low + score_high
    if len(scores) == 0:
        return constrain(thresholds[0])
    index = np.argwhere(scores == scores.min()).flatten()
    if len(index) == 0:
        return constrain(thresholds[0])
    #
    # Take the average of the thresholds to either side of
    # the chosen value to get an intermediate in cases where there is
    # a steep step between the background and foreground
    index = index[0]
    if index == 0:
        index_low = 0
    else:
        index_low = index - 1
    if index == len(thresholds) - 1:
        index_high = len(thresholds) - 1
    else:
        index_high = index + 1
    return constrain((thresholds[index_low] + thresholds[index_high]) / 2)


def entropy(data, bins=256):
    """Compute a threshold using Ray's entropy measurement
    
    data           - an array of intensity values between zero and one
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    """

    data = np.atleast_1d(data)
    data = data[~np.isnan(data)]
    if len(data) == 0:
        return 0
    elif len(data) == 1:
        return data[0]

    if bins > len(data):
        bins = len(data)
    data.sort()
    var = running_variance(data) + 1.0 / 512.0
    rvar = np.flipud(running_variance(np.flipud(data))) + 1.0 / 512.0
    thresholds = data[1 : len(data) : len(data) // bins]
    w = np.arange(0, len(data) - 1, len(data) // bins)
    score_low = w * np.log(
        var[0 : len(data) - 1 : len(data) // bins] * w * np.sqrt(2 * np.pi * np.exp(1))
    )
    score_low[np.isnan(score_low)] = 0

    w = len(data) - np.arange(1, len(data), len(data) // bins)
    score_high = w * np.log(
        rvar[1 : len(data) : len(data) // bins] * w * np.sqrt(2 * np.pi * np.exp(1))
    )
    score_high[np.isnan(score_high)] = 0
    scores = score_low + score_high
    index = np.argwhere(scores == scores.min()).flatten()
    if len(index) == 0:
        return thresholds[0]
    #
    # Take the average of the thresholds to either side of
    # the chosen value to get an intermediate in cases where there is
    # a steep step between the background and foreground
    index = index[0]
    if index == 0:
        index_low = 0
    else:
        index_low = index - 1
    if index == len(thresholds) - 1:
        index_high = len(thresholds) - 1
    else:
        index_high = index + 1
    return (thresholds[index_low] + thresholds[index_high]) / 2


def otsu3(data, min_threshold=None, max_threshold=None, bins=128):
    """Compute a threshold using a 3-category Otsu-like method
    
    data           - an array of intensity values between zero and one
    min_threshold  - only consider thresholds above this minimum value
    max_threshold  - only consider thresholds below this maximum value
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    
    We find the maximum weighted variance, breaking the histogram into
    three pieces.
    Returns the lower and upper thresholds
    """
    assert (
        min_threshold is None or max_threshold is None or min_threshold < max_threshold
    )

    #
    # Compute the running variance and reverse running variance.
    #
    data = np.atleast_1d(data)
    data = data[~np.isnan(data)]
    data.sort()
    if len(data) == 0:
        return 0
    var = running_variance(data)
    rvar = np.flipud(running_variance(np.flipud(data)))
    if bins > len(data):
        bins = len(data)
    bin_len = int(len(data) // bins)
    thresholds = data[0 : len(data) : bin_len]
    score_low = var[0 : len(data) : bin_len] * np.arange(0, len(data), bin_len)
    score_high = rvar[0 : len(data) : bin_len] * (
        len(data) - np.arange(0, len(data), bin_len)
    )
    #
    # Compute the middles
    #
    cs = data.cumsum()
    cs2 = (data ** 2).cumsum()
    i, j = np.mgrid[0 : score_low.shape[0], 0 : score_high.shape[0]] * bin_len
    diff = (j - i).astype(float)
    w = diff
    mean = (cs[j] - cs[i]) / diff
    mean2 = (cs2[j] - cs2[i]) / diff
    score_middle = w * (mean2 - mean ** 2)
    score_middle[i >= j] = np.Inf
    score = (
        score_low[i * bins // len(data)]
        + score_middle
        + score_high[j * bins // len(data)]
    )
    best_score = np.min(score)
    best_i_j = np.argwhere(score == best_score)
    return (thresholds[best_i_j[0, 0]], thresholds[best_i_j[0, 1]])


def entropy3(data, bins=128):
    """Compute a threshold using a 3-category Otsu-like method
    
    data           - an array of intensity values between zero and one
    bins           - we bin the data into this many equally-spaced bins, then pick
                     the bin index that optimizes the metric
    
    We find the maximum weighted variance, breaking the histogram into
    three pieces.
    Returns the lower and upper thresholds
    """
    #
    # Compute the running variance and reverse running variance.
    #
    data = np.atleast_1d(data)
    data = data[~np.isnan(data)]
    data.sort()
    if len(data) == 0:
        return 0
    var = running_variance(data) + 1.0 / 512.0
    if bins > len(data):
        bins = len(data)
    bin_len = int(len(data) // bins)
    thresholds = data[0 : len(data) : bin_len]
    score_low = entropy_score(var, bins)

    rvar = running_variance(np.flipud(data)) + 1.0 / 512.0
    score_high = np.flipud(entropy_score(rvar, bins))
    #
    # Compute the middles
    #
    cs = data.cumsum()
    cs2 = (data ** 2).cumsum()
    i, j = np.mgrid[0 : score_low.shape[0], 0 : score_high.shape[0]] * bin_len
    diff = (j - i).astype(float)
    w = diff / float(len(data))
    mean = (cs[j] - cs[i]) / diff
    mean2 = (cs2[j] - cs2[i]) / diff
    score_middle = entropy_score(mean2 - mean ** 2 + 1.0 / 512.0, bins, w, False)
    score_middle[(i >= j) | np.isnan(score_middle)] = np.Inf
    score = score_low[i // bin_len] + score_middle + score_high[j // bin_len]
    best_score = np.min(score)
    best_i_j = np.argwhere(score == best_score)
    return (thresholds[best_i_j[0, 0]], thresholds[best_i_j[0, 1]])


def entropy_score(var, bins, w=None, decimate=True):
    """Compute entropy scores, given a variance and # of bins
    
    """
    if w is None:
        n = len(var)
        w = np.arange(0, n, n // bins) / float(n)
    if decimate:
        n = len(var)
        var = var[0 : n : n // bins]
    score = w * np.log(var * w * np.sqrt(2 * np.pi * np.exp(1)))
    score[np.isnan(score)] = np.Inf
    return score


def weighted_variance(cs, cs2, lo, hi):
    if hi == lo:
        return np.Infinity
    w = (hi - lo) / float(len(cs))
    mean = (cs[hi] - cs[lo]) / (hi - lo)
    mean2 = (cs2[hi] - cs2[lo]) / (hi - lo)
    return w * (mean2 - mean ** 2)


def otsu_entropy(cs, cs2, lo, hi):
    if hi == lo:
        return np.Infinity
    w = (hi - lo) / float(len(cs))
    mean = (cs[hi] - cs[lo]) / (hi - lo)
    mean2 = (cs2[hi] - cs2[lo]) / (hi - lo)
    return w * (np.log(w * (mean2 - mean ** 2) * np.sqrt(2 * np.pi * np.exp(1))))


def running_variance(x):
    """Given a vector x, compute the variance for x[0:i]
    
    Thank you http://www.johndcook.com/standard_deviation.html
    S[i] = S[i-1]+(x[i]-mean[i-1])*(x[i]-mean[i])
    var(i) = S[i] / (i-1)
    """
    n = len(x)
    # The mean of x[0:i]
    m = x.cumsum() / np.arange(1, n + 1)
    # x[i]-mean[i-1] for i=1...
    x_minus_mprev = x[1:] - m[:-1]
    # x[i]-mean[i] for i=1...
    x_minus_m = x[1:] - m[1:]
    # s for i=1...
    s = (x_minus_mprev * x_minus_m).cumsum()
    var = s / np.arange(1, n)
    # Prepend Inf so we have a variance for x[0]
    return np.hstack(([0], var))



from __future__ import absolute_import
import glob
import os.path
import sys

import pkg_resources
import setuptools
import setuptools.command.build_ext
import setuptools.command.test

try:
    import Cython.Build

    __cython = True
except ImportError:
    __cython = False


class BuildExtension(setuptools.command.build_ext.build_ext):
    def build_extensions(self):
        numpy_includes = pkg_resources.resource_filename("numpy", "core/include")

        for extension in self.extensions:
            if (
                hasattr(extension, "include_dirs")
                and numpy_includes not in extension.include_dirs
            ):
                extension.include_dirs.append(numpy_includes)

            extension.include_dirs.append("centrosome/include")

        setuptools.command.build_ext.build_ext.build_extensions(self)


class Test(setuptools.command.test.test):
    user_options = [("pytest-args=", "a", "Arguments to pass to py.test")]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)

        self.pytest_args = []

    def finalize_options(self):
        setuptools.command.test.test.finalize_options(self)

        self.test_args = []

        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)

        sys.exit(errno)


if __cython:
    __suffix = "pyx"
    __extkwargs = {"language": "c++"}
else:
    __suffix = "cpp"
    __extkwargs = {}

__extensions = [
    setuptools.Extension(
        name="centrosome._propagate",
        sources=[
            "centrosome/_propagate.{}".format("c" if __suffix == "cpp" else __suffix)
        ],
    )
]

for pyxfile in glob.glob(os.path.join("centrosome", "*.pyx")):
    name = os.path.splitext(os.path.basename(pyxfile))[0]

    if name == "_propagate":
        continue

    __extensions += [
        setuptools.Extension(
            name="centrosome.{}".format(name),
            sources=["centrosome/{}.{}".format(name, __suffix)],
            **__extkwargs
        )
    ]

if __suffix == "pyx":
    __extensions = Cython.Build.cythonize(__extensions)

setuptools.setup(
    author="Allen Goodman",
    author_email="agoodman@broadinstitute.org",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={"build_ext": BuildExtension, "test": Test},
    description="An open source image processing library",
    ext_modules=__extensions,
    extras_require={
        "dev": ["black==19.10b0", "pre-commit==1.20.0"],
        "test": ["pytest==5.2.2"],
    },
    install_requires=[
        "deprecation",
        "matplotlib==3.1.3",
        "numpy>=1.18.2",
        "pillow>=7.1.0",
        "scikit-image>=0.17.2",
        "scipy>=1.4.1",
    ],
    tests_require=[
        "pytest",
    ],
    keywords="",
    license="BSD",
    long_description="",
    name="centrosome",
    packages=["centrosome"],
    setup_requires=["cython", "numpy", "pytest",],
    url="https://github.com/CellProfiler/centrosome",
    version="1.2.0",
)



from __future__ import absolute_import
import numpy


def outline(labels):
    """Given a label matrix, return a matrix of the outlines of the labeled objects
    
    If a pixel is not zero and has at least one neighbor with a different
    value, then it is part of the outline.
    """

    output = numpy.zeros(labels.shape, labels.dtype)
    lr_different = labels[1:, :] != labels[:-1, :]
    ud_different = labels[:, 1:] != labels[:, :-1]
    d1_different = labels[1:, 1:] != labels[:-1, :-1]
    d2_different = labels[1:, :-1] != labels[:-1, 1:]
    different = numpy.zeros(labels.shape, bool)
    different[1:, :][lr_different] = True
    different[:-1, :][lr_different] = True
    different[:, 1:][ud_different] = True
    different[:, :-1][ud_different] = True
    different[1:, 1:][d1_different] = True
    different[:-1, :-1][d1_different] = True
    different[1:, :-1][d2_different] = True
    different[:-1, 1:][d2_different] = True
    #
    # Labels on edges need outlines
    #
    different[0, :] = True
    different[:, 0] = True
    different[-1, :] = True
    different[:, -1] = True

    output[different] = labels[different]
    return output
